class TemplateVariables:
    """The names of template variables."""
    CODE_GENERATION_PROMPT = 'code_generation_prompt'
    CONTEXT = 'context'
    PROMPT = 'prompt'
    CONVERSATION_FILES = 'conversation_files'
    ATTACHED_FILES = 'attached_files'
    CONTEXT_FILES = 'context_files'
    TOOL_RESULTS = 'tool_results'
    FILES_PROMPT = 'files_prompt'
